@@ -39,7 +39,14 @@
 #include <arpa/inet.h>		/* for htonl, ntohl */
 
 #ifdef linux
-#include <malloc.h>
+#  include <malloc.h>
+
+#  define OOM_ADJ_FILE "/proc/self/oom_adj"
+
+/* From linux/oom.h */
+#  define OOM_DISABLE (-17)
+#  define OOM_ADJUST_MIN (-16)
+
 #endif
 
 /* FIXME We use syslog for now, because multilog is not yet implemented */
@@ -47,6 +54,7 @@
 
 static volatile sig_atomic_t _exit_now = 0;	/* set to '1' when signal is given to exit */
 static volatile sig_atomic_t _thread_registries_empty = 1;	/* registries are empty initially */
+static int _debug = 0;
 
 /* List (un)link macros. */
 #define	LINK(x, head)		list_add(head, &(x)->list)
@@ -60,7 +68,7 @@ static volatile sig_atomic_t _thread_registries_empty = 1;	/* registries are emp
 #define DAEMON_NAME "dmeventd"
 
 /*
-  Global mutex for thread list access. Has to be held when: 
+  Global mutex for thread list access. Has to be held when:
   - iterating thread list
   - adding or removing elements from thread list
   - changing or reading thread_status's fields:
@@ -69,12 +77,27 @@ static volatile sig_atomic_t _thread_registries_empty = 1;	/* registries are emp
 */
 static pthread_mutex_t _global_mutex;
 
+/*
+  There are three states a thread can attain (see struct
+  thread_status, field int status):
+
+  - DM_THREAD_RUNNING: thread has started up and is either working or
+  waiting for events... transitions to either SHUTDOWN or DONE
+  - DM_THREAD_SHUTDOWN: thread is still doing something, but it is
+  supposed to terminate (and transition to DONE) as soon as it
+  finishes whatever it was doing at the point of flipping state to
+  SHUTDOWN... the thread is still on the thread list
+  - DM_THREAD_DONE: thread has terminated and has been moved over to
+  unused thread list, cleanup pending
+ */
 #define DM_THREAD_RUNNING  0
 #define DM_THREAD_SHUTDOWN 1
 #define DM_THREAD_DONE     2
 
 #define THREAD_STACK_SIZE (300*1024)
 
+#define DEBUGLOG(fmt, args...) _debuglog(fmt, ## args)
+
 /* Data kept about a DSO. */
 struct dso_data {
 	struct list list;
@@ -96,7 +119,7 @@ struct dso_data {
 	 * DM_DEVICE_STATUS). It should not destroy it.
 	 * The caller must dispose of the task.
 	 */
-	void (*process_event)(struct dm_task *dmt, enum dm_event_mask event);
+	void (*process_event)(struct dm_task *dmt, enum dm_event_mask event, void **user);
 
 	/*
 	 * Device registration.
@@ -107,7 +130,7 @@ struct dso_data {
 	 * and activate a mapping).
 	 */
 	int (*register_device)(const char *device, const char *uuid, int major,
-			       int minor);
+			       int minor, void **user);
 
 	/*
 	 * Device unregistration.
@@ -117,7 +140,7 @@ struct dso_data {
 	 * steps (eg, deactivate mapping, metadata update).
 	 */
 	int (*unregister_device)(const char *device, const char *uuid,
-				 int major, int minor);
+				 int major, int minor, void **user);
 };
 static LIST_INIT(_dso_registry);
 
@@ -156,13 +179,16 @@ struct thread_status {
 	} device;
 	uint32_t event_nr;	/* event number */
 	int processing;		/* Set when event is being processed */
-	int status;		/* running/shutdown/done */
+
+	int status;		/* see DM_THREAD_{RUNNING,SHUTDOWN,DONE}
+				   constants above */
 	enum dm_event_mask events;	/* bitfield for event filter. */
 	enum dm_event_mask current_events;	/* bitfield for occured events. */
 	struct dm_task *current_task;
 	time_t next_time;
 	uint32_t timeout;
 	struct list timeout_list;
+	void *dso_private; /* dso per-thread status variable */
 };
 static LIST_INIT(_thread_registry);
 static LIST_INIT(_thread_registry_unused);
@@ -172,6 +198,24 @@ static LIST_INIT(_timeout_registry);
 static pthread_mutex_t _timeout_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t _timeout_cond = PTHREAD_COND_INITIALIZER;
 
+static void _debuglog(const char *fmt, ...)
+{
+        time_t P;
+        va_list ap;
+ 
+        if (!_debug)
+                return;
+ 
+        va_start(ap,fmt);
+
+        time(&P);
+        fprintf(stderr, "dmeventd[%p]: %.15s ", (void *) pthread_self(), ctime(&P)+4 );
+        vfprintf(stderr, fmt, ap);
+	fprintf(stderr, "\n");
+
+        va_end(ap);
+}
+
 /* Allocate/free the status structure for a monitoring thread. */
 static struct thread_status *_alloc_thread_status(struct message_data *data,
 						  struct dso_data *dso_data)
@@ -422,14 +466,14 @@ static void _exit_dm_lib(void)
 	dm_lib_exit();
 }
 
-static void _exit_timeout(void *unused)
+static void _exit_timeout(void *unused __attribute((unused)))
 {
 	_timeout_running = 0;
 	pthread_mutex_unlock(&_timeout_mutex);
 }
 
 /* Wake up monitor threads every so often. */
-static void *_timeout_thread(void *unused)
+static void *_timeout_thread(void *unused __attribute((unused)))
 {
 	struct timespec timeout;
 	time_t curr_time;
@@ -441,16 +485,16 @@ static void *_timeout_thread(void *unused)
 	while (!list_empty(&_timeout_registry)) {
 		struct thread_status *thread;
 
-		timeout.tv_sec = (time_t) -1;
+		timeout.tv_sec = 0;
 		curr_time = time(NULL);
 
 		list_iterate_items_gen(thread, &_timeout_registry, timeout_list) {
-			if (thread->next_time < curr_time) {
+			if (thread->next_time <= curr_time) {
 				thread->next_time = curr_time + thread->timeout;
 				pthread_kill(thread->thread, SIGALRM);
 			}
 
-			if (thread->next_time < timeout.tv_sec)
+			if (thread->next_time < timeout.tv_sec || !timeout.tv_sec)
 				timeout.tv_sec = thread->next_time;
 		}
 
@@ -602,7 +646,8 @@ static int _do_register_device(struct thread_status *thread)
 	return thread->dso_data->register_device(thread->device.name,
 						 thread->device.uuid,
 						 thread->device.major,
-						 thread->device.minor);
+						 thread->device.minor,
+						 &(thread->dso_private));
 }
 
 /* Unregister a device with the DSO. */
@@ -611,13 +656,14 @@ static int _do_unregister_device(struct thread_status *thread)
 	return thread->dso_data->unregister_device(thread->device.name,
 						   thread->device.uuid,
 						   thread->device.major,
-						   thread->device.minor);
+						   thread->device.minor,
+						   &(thread->dso_private));
 }
 
 /* Process an event in the DSO. */
 static void _do_process_event(struct thread_status *thread, struct dm_task *task)
 {
-	thread->dso_data->process_event(task, thread->current_events);
+	thread->dso_data->process_event(task, thread->current_events, &(thread->dso_private));
 }
 
 /* Thread cleanup handler to unregister device. */
@@ -655,6 +701,23 @@ static void _monitor_unregister(void *arg)
 	_unlock_mutex();
 }
 
+static struct dm_task *_get_device_status(struct thread_status *ts)
+{
+	struct dm_task *dmt = dm_task_create(DM_DEVICE_STATUS);
+
+	if (!dmt)
+		return NULL;
+
+	dm_task_set_uuid(dmt, ts->device.uuid);
+
+	if (!dm_task_run(dmt)) {
+		dm_task_destroy(dmt);
+		return NULL;
+	}
+
+	return dmt;
+}
+
 /* Device monitoring thread. */
 static void *_monitor_thread(void *arg)
 {
@@ -681,6 +744,18 @@ static void *_monitor_thread(void *arg)
 		if (wait_error == DM_WAIT_FATAL)
 			break;
 
+		/* Timeout occurred, task is not filled properly.
+		 * We get device status here for processing it in DSO.
+		 */
+		if (wait_error == DM_WAIT_INTR &&
+		    thread->current_events & DM_EVENT_TIMEOUT) {
+			dm_task_destroy(task);
+			task = _get_device_status(thread);
+			/* FIXME: syslog fail here ? */
+			if (!(thread->current_task = task))
+				continue;
+		}
+
 		/*
 		 * We know that wait succeeded and stored a
 		 * pointer to dm_task with device status into task.
@@ -771,8 +846,7 @@ static struct dso_data *_lookup_dso(struct message_data *data)
 }
 
 /* Lookup DSO symbols we need. */
-static int _lookup_symbol(void *dl, struct dso_data *data,
-			 void **symbol, const char *name)
+static int _lookup_symbol(void *dl, void **symbol, const char *name)
 {
 	if ((*symbol = dlsym(dl, name)))
 		return 1;
@@ -782,11 +856,11 @@ static int _lookup_symbol(void *dl, struct dso_data *data,
 
 static int lookup_symbols(void *dl, struct dso_data *data)
 {
-	return _lookup_symbol(dl, data, (void *) &data->process_event,
+	return _lookup_symbol(dl, (void *) &data->process_event,
 			     "process_event") &&
-	    _lookup_symbol(dl, data, (void *) &data->register_device,
+	    _lookup_symbol(dl, (void *) &data->register_device,
 			  "register_device") &&
-	    _lookup_symbol(dl, data, (void *) &data->unregister_device,
+	    _lookup_symbol(dl, (void *) &data->unregister_device,
 			  "unregister_device");
 }
 
@@ -882,12 +956,12 @@ static int _register_for_event(struct message_data *message_data)
 	   usually means we are so starved on resources that we are
 	   almost as good as dead already... */
 	if (thread_new->events & DM_EVENT_TIMEOUT) {
-		ret = -_register_for_timeout(thread);
-                if (ret) {
-                    _unlock_mutex();
-                    goto out;
-                }
-        }
+		ret = -_register_for_timeout(thread_new);
+		if (ret) {
+		    _unlock_mutex();
+		    goto out;
+		}
+	}
 
 	if (!(thread = _lookup_thread_status(message_data))) {
 		_unlock_mutex();
@@ -1051,22 +1125,25 @@ static int _get_registered_dev(struct message_data *message_data, int next)
 	if (!hit)
 		goto out;
 
-	goto out; /* FIXME the next == 1 thing is currently horridly
-		     broken, do something about it... */
+	thread = hit;
 
-	do {
+	while (1) {
 		if (list_end(&_thread_registry, &thread->list))
 			goto out;
 
 		thread = list_item(thread->list.n, struct thread_status);
-	} while (!_want_registered_device(message_data->dso_name, NULL, thread));
+		if (_want_registered_device(message_data->dso_name, NULL, thread)) {
+			hit = thread;
+			break;
+		}
+	}
 
 	_unlock_mutex();
-	return _registered_device(message_data, thread);
+	return _registered_device(message_data, hit);
 
       out:
 	_unlock_mutex();
-
+	
 	return -ENOENT;
 }
 
@@ -1361,7 +1438,7 @@ static void _cleanup_unused_threads(void)
 		if (thread->status == DM_THREAD_RUNNING) {
 			thread->status = DM_THREAD_SHUTDOWN;
 			break;
-		} 
+		}
 
 		if (thread->status == DM_THREAD_SHUTDOWN) {
 			if (!thread->events) {
@@ -1377,7 +1454,7 @@ static void _cleanup_unused_threads(void)
 					stack;
 				}
 				break;
-			} 
+			}
 
 			list_del(l);
 			syslog(LOG_ERR,
@@ -1399,7 +1476,7 @@ static void _cleanup_unused_threads(void)
 	_unlock_mutex();
 }
 
-static void _sig_alarm(int signum)
+static void _sig_alarm(int signum __attribute((unused)))
 {
 	pthread_testcancel();
 }
@@ -1431,7 +1508,7 @@ static void _init_thread_signals(void)
  * Set the global variable which the process should
  * be watching to determine when to exit.
  */
-static void _exit_handler(int sig)
+static void _exit_handler(int sig __attribute((unused)))
 {
 	/*
 	 * We exit when '_exit_now' is set.
@@ -1465,50 +1542,64 @@ static int _lock_pidfile(void)
 	return 0;
 }
 
+#ifdef linux
+/*
+ * Protection against OOM killer if kernel supports it
+ */
 static int _set_oom_adj(int val)
 {
 	FILE *fp;
 
 	struct stat st;
 
-	if (stat("/proc/self/oom_adj", &st) == -1)
-		return -errno;
+	if (stat(OOM_ADJ_FILE, &st) == -1) {
+		if (errno == ENOENT)
+			DEBUGLOG(OOM_ADJ_FILE " not found");
+		else
+			perror(OOM_ADJ_FILE ": stat failed");
+		return 1;
+	}
 
-	fp = fopen("/proc/self/oom_adj", "w");
-
-	if (!fp)
-		return -1;
+	if (!(fp = fopen(OOM_ADJ_FILE, "w"))) {
+		perror(OOM_ADJ_FILE ": fopen failed");
+		return 0;
+	}
 
 	fprintf(fp, "%i", val);
-	fclose(fp);
+	if (fclose(fp))
+		perror(OOM_ADJ_FILE ": fclose failed");
 
-	return 0;
+	return 1;
 }
+#endif
 
 static void _daemonize(void)
 {
-	int status;
-	int pid;
+	int child_status;
 	int fd;
+	pid_t pid;
 	struct rlimit rlim;
 	struct timeval tval;
 	sigset_t my_sigset;
 
 	sigemptyset(&my_sigset);
 	if (sigprocmask(SIG_SETMASK, &my_sigset, NULL) < 0) {
-		fprintf(stderr, "Unable to restore signals.");
+		fprintf(stderr, "Unable to restore signals.\n");
 		exit(EXIT_FAILURE);
 	}
 	signal(SIGTERM, &_exit_handler);
 
-	pid = fork();
-
-	if (pid < 0)
+	switch (pid = fork()) {
+	case -1:
+		perror("fork failed:");
 		exit(EXIT_FAILURE);
 
-	if (pid) {
+	case 0:		/* Child */
+		break;
+
+	default:
 		/* Wait for response from child */
-		while (!waitpid(pid, &status, WNOHANG) && !_exit_now) {
+		while (!waitpid(pid, &child_status, WNOHANG) && !_exit_now) {
 			tval.tv_sec = 0;
 			tval.tv_usec = 250000;	/* .25 sec */
 			select(0, NULL, NULL, NULL, &tval);
@@ -1518,27 +1609,23 @@ static void _daemonize(void)
 			exit(EXIT_SUCCESS);
 
 		/* Problem with child.  Determine what it is by exit code */
-		switch (WEXITSTATUS(status)) {
+		switch (WEXITSTATUS(child_status)) {
 		case EXIT_LOCKFILE_INUSE:
+			fprintf(stderr, "Another dmeventd daemon is already running\n");
 			break;
 		case EXIT_DESC_CLOSE_FAILURE:
-			break;
 		case EXIT_DESC_OPEN_FAILURE:
-			break;
 		case EXIT_OPEN_PID_FAILURE:
-			break;
 		case EXIT_FIFO_FAILURE:
-			break;
 		case EXIT_CHDIR_FAILURE:
-			break;
 		default:
+			fprintf(stderr, "Child exited with code %d\n", WEXITSTATUS(child_status));
 			break;
 		}
 
-		exit(EXIT_FAILURE);	/* Redundant */
+		exit(WEXITSTATUS(child_status));
 	}
 
-	setsid();
 	if (chdir("/"))
 		exit(EXIT_CHDIR_FAILURE);
 
@@ -1555,6 +1642,51 @@ static void _daemonize(void)
 	    (open("/dev/null", O_WRONLY) < 0))
 		exit(EXIT_DESC_OPEN_FAILURE);
 
+	setsid();
+}
+
+static void usage(char *prog, FILE *file)
+{
+	fprintf(file, "Usage:\n");
+	fprintf(file, "%s [Vhd]\n", prog);
+	fprintf(file, "\n");
+	fprintf(file, "   -V       Show version of dmeventd\n");
+	fprintf(file, "   -h       Show this help information\n");
+	fprintf(file, "   -d       Don't fork, run in the foreground\n");
+	fprintf(file, "\n");
+}
+
+int main(int argc, char *argv[])
+{
+	int ret;
+	signed char opt;
+	struct dm_event_fifos fifos;
+	//struct sys_log logdata = {DAEMON_NAME, LOG_DAEMON};
+
+	opterr = 0;
+	optind = 0;
+
+	while ((opt = getopt(argc, argv, "?hVd")) != EOF) {
+		switch (opt) {
+		case 'h':
+			usage(argv[0], stdout);
+			exit(0);
+		case '?':
+			usage(argv[0], stderr);
+			exit(0);
+		case 'd':
+			_debug++;
+			break;
+		case 'V':
+			printf("dmeventd version: %s\n", DM_LIB_VERSION);
+			exit(1);
+			break;
+		}
+	}
+
+	if (!_debug)
+		_daemonize();
+
 	openlog("dmeventd", LOG_PID, LOG_DAEMON);
 
 	_lock_pidfile();		/* exits if failure */
@@ -1563,21 +1695,11 @@ static void _daemonize(void)
 	signal(SIGINT, &_exit_handler);
 	signal(SIGHUP, &_exit_handler);
 	signal(SIGQUIT, &_exit_handler);
-}
 
-int main(int argc, char *argv[])
-{
-	int ret;
-	struct dm_event_fifos fifos;
-	//struct sys_log logdata = {DAEMON_NAME, LOG_DAEMON};
-
-	_daemonize();
-
-	/*
-	 * ENOENT means the kernel does not support oom_adj
-	 */
-	if (_set_oom_adj(-16) != -ENOENT)
+#ifdef linux
+	if (!_set_oom_adj(OOM_DISABLE) && !_set_oom_adj(OOM_ADJUST_MIN))
 		syslog(LOG_ERR, "Failed to set oom_adj to protect against OOM killer");
+#endif
 
 	_init_thread_signals();
 
@@ -1600,7 +1722,7 @@ int main(int argc, char *argv[])
 
 	/* Signal parent, letting them know we are ready to go. */
 	kill(getppid(), SIGTERM);
-	syslog(LOG_INFO, "dmeventd ready for processing.");
+	syslog(LOG_NOTICE, "dmeventd ready for processing.");
 
 	while (!_exit_now) {
 		_process_request(&fifos);
@@ -1619,7 +1741,8 @@ int main(int argc, char *argv[])
 #endif
 	pthread_mutex_destroy(&_global_mutex);
 
-	syslog(LOG_INFO, "dmeventd shutting down.");
+	syslog(LOG_NOTICE, "dmeventd shutting down.");
 	closelog();
+
 	exit(EXIT_SUCCESS);
 }
